{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "A small notebook explaining the pitfalls of algorithmic compexities and the difference of complexity and actual runtime."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "First, let's define a few mockup functions whose performance we want to analyze."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "# mockup for a function with linear runtime:\n",
    "# it simply performs a multiplication on the input parameters\n",
    "# this function has constant runtime,\n",
    "# basically just the costs for one multiplication\n",
    "def foo(i,j):\n",
    "    return i*j\n",
    "\n",
    "# mockup for a function with linear runtime:\n",
    "# this function calls function <func> n times\n",
    "def loop(n, func):\n",
    "    sum = 0\n",
    "    for i in range(n):\n",
    "        sum += func(i,i)\n",
    "    return sum\n",
    "\n",
    "# mockup for a function with quadratic runtime:\n",
    "# this function calls function <func> n*n times\n",
    "def nestedLoop(n, func):\n",
    "    sum = 0\n",
    "    for i in range(n):\n",
    "        for j in range(n):\n",
    "            sum += func(i,j)\n",
    "    return sum\n",
    "\n",
    "# mockup for a function with n*log n runtime:\n",
    "# this function calls function <func> n*ceil(log_2 n) times\n",
    "import math\n",
    "def loopLog(n, func):\n",
    "    sum = 0\n",
    "    _log = math.ceil(math.log2(n))\n",
    "    for i in range(n):\n",
    "        for j in range(_log):\n",
    "            sum += func(i,j)\n",
    "    return sum"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[('23 Databases.pdf', 100938180), ('24 Database Indexing.pdf', 9474897), ('geonamesGermany.db', 19877888), ('geonamesWorld.db', 3217985536), ('24 Databases 2.pdf', 95571426), ('slide collection.pdf', 19056675), ('geonames.db', 13127680)]\n"
     ]
    }
   ],
   "source": [
    "# list files in current directory greater than 1MB\n",
    "# we do this to simulate I/O operations to play with:\n",
    "import os\n",
    "files = [(f,os.stat(f).st_size) for f in os.listdir('.') if os.path.isfile(f) and os.stat(f).st_size >1024*1024]\n",
    "print(files)\n",
    "\n",
    "# a mockup function performing a pseudo-random read to simulate I/O:\n",
    "def foo_IO(i,j):\n",
    "    # pick a file pseudo-randomly from the pre-computed list of files:\n",
    "    pseudo_random = i*i % len(files)\n",
    "    with open(files[pseudo_random][0], 'rb') as f:\n",
    "        # seek to random offset in that file:\n",
    "        f.seek(files[pseudo_random][1]//2)\n",
    "        # read 4096 bytes and return the last byte read:\n",
    "        first_bytes = f.read(4096)\n",
    "        # return the bytes read:\n",
    "        return first_bytes[min(4095,len(first_bytes)-1)]"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Let's run performance measurements on these mockup functions:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "metadata": {
    "scrolled": false
   },
   "outputs": [
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "100%|██████████| 4/4 [02:30<00:00, 37.74s/it]\n"
     ]
    }
   ],
   "source": [
    "import time\n",
    "from tqdm import tqdm\n",
    "# combination of methods to benchmark:\n",
    "# syntax: (outer-function, inner-function)\n",
    "methods=[ (loop,foo), (loop, foo_IO), (loopLog, foo), (nestedLoop, foo)]\n",
    "results = []\n",
    "for m in tqdm(methods):\n",
    "    # vary n:\n",
    "    for n in range(100, 10000, 500):\n",
    "        start = time.time()\n",
    "        # call the function combination:\n",
    "        m[0](n, m[1])\n",
    "        elapsedTime = time.time() - start\n",
    "        # append measured times to measurement data:\n",
    "        results.append((m[0].__name__, m[1].__name__, n, elapsedTime))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Let's display the performance results visually:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "\n",
       "<div id=\"altair-viz-4\"></div>\n",
       "<script type=\"text/javascript\">\n",
       "  (function(spec, embedOpt){\n",
       "    const outputDiv = document.getElementById(\"altair-viz-4\");\n",
       "    const paths = {\n",
       "      \"vega\": \"https://cdn.jsdelivr.net/npm//vega@5?noext\",\n",
       "      \"vega-lib\": \"https://cdn.jsdelivr.net/npm//vega-lib?noext\",\n",
       "      \"vega-lite\": \"https://cdn.jsdelivr.net/npm//vega-lite@4.0.0?noext\",\n",
       "      \"vega-embed\": \"https://cdn.jsdelivr.net/npm//vega-embed@6?noext\",\n",
       "    };\n",
       "\n",
       "    function loadScript(lib) {\n",
       "      return new Promise(function(resolve, reject) {\n",
       "        var s = document.createElement('script');\n",
       "        s.src = paths[lib];\n",
       "        s.async = true;\n",
       "        s.onload = () => resolve(paths[lib]);\n",
       "        s.onerror = () => reject(`Error loading script: ${paths[lib]}`);\n",
       "        document.getElementsByTagName(\"head\")[0].appendChild(s);\n",
       "      });\n",
       "    }\n",
       "\n",
       "    function showError(err) {\n",
       "      outputDiv.innerHTML = `<div class=\"error\" style=\"color:red;\">${err}</div>`;\n",
       "      throw err;\n",
       "    }\n",
       "\n",
       "    function displayChart(vegaEmbed) {\n",
       "      vegaEmbed(outputDiv, spec, embedOpt)\n",
       "        .catch(err => showError(`Javascript Error: ${err.message}<br>This usually means there's a typo in your chart specification. See the javascript console for the full traceback.`));\n",
       "    }\n",
       "\n",
       "    if(typeof define === \"function\" && define.amd) {\n",
       "      requirejs.config({paths});\n",
       "      require([\"vega-embed\"], displayChart, err => showError(`Error loading script: ${err.message}`));\n",
       "    } else if (typeof vegaEmbed === \"function\") {\n",
       "      displayChart(vegaEmbed);\n",
       "    } else {\n",
       "      loadScript(\"vega\")\n",
       "        .then(() => loadScript(\"vega-lite\"))\n",
       "        .then(() => loadScript(\"vega-embed\"))\n",
       "        .catch(showError)\n",
       "        .then(() => displayChart(vegaEmbed));\n",
       "    }\n",
       "  })({\"config\": {\"view\": {\"continuousWidth\": 400, \"continuousHeight\": 300}}, \"data\": {\"name\": \"data-53ce784c6655b5a555c246ab8f8f0958\"}, \"mark\": \"line\", \"encoding\": {\"color\": {\"type\": \"nominal\", \"field\": \"cat\", \"title\": \"Algorithm\"}, \"x\": {\"type\": \"quantitative\", \"field\": \"n\"}, \"y\": {\"type\": \"quantitative\", \"field\": \"time\", \"title\": \"Time [sec]\"}}, \"selection\": {\"selector003\": {\"type\": \"interval\", \"bind\": \"scales\", \"encodings\": [\"x\", \"y\"]}}, \"transform\": [{\"calculate\": \"datum.method + ' using ' + datum.foo\", \"as\": \"cat\"}], \"$schema\": \"https://vega.github.io/schema/vega-lite/v4.0.0.json\", \"datasets\": {\"data-53ce784c6655b5a555c246ab8f8f0958\": [{\"method\": \"loop\", \"foo\": \"foo\", \"n\": 100, \"time\": 2.193450927734375e-05}, {\"method\": \"loop\", \"foo\": \"foo\", \"n\": 600, \"time\": 0.00011491775512695312}, {\"method\": \"loop\", \"foo\": \"foo\", \"n\": 1100, \"time\": 0.00021195411682128906}, {\"method\": \"loop\", \"foo\": \"foo\", \"n\": 1600, \"time\": 0.0003287792205810547}, {\"method\": \"loop\", \"foo\": \"foo\", \"n\": 2100, \"time\": 0.0004558563232421875}, {\"method\": \"loop\", \"foo\": \"foo\", \"n\": 2600, \"time\": 0.0006310939788818359}, {\"method\": \"loop\", \"foo\": \"foo\", \"n\": 3100, \"time\": 0.0007102489471435547}, {\"method\": \"loop\", \"foo\": \"foo\", \"n\": 3600, \"time\": 0.0010559558868408203}, {\"method\": \"loop\", \"foo\": \"foo\", \"n\": 4100, \"time\": 0.0009198188781738281}, {\"method\": \"loop\", \"foo\": \"foo\", \"n\": 4600, \"time\": 0.00098419189453125}, {\"method\": \"loop\", \"foo\": \"foo\", \"n\": 5100, \"time\": 0.0010759830474853516}, {\"method\": \"loop\", \"foo\": \"foo\", \"n\": 5600, \"time\": 0.0012178421020507812}, {\"method\": \"loop\", \"foo\": \"foo\", \"n\": 6100, \"time\": 0.0012972354888916016}, {\"method\": \"loop\", \"foo\": \"foo\", \"n\": 6600, \"time\": 0.0015377998352050781}, {\"method\": \"loop\", \"foo\": \"foo\", \"n\": 7100, \"time\": 0.0017008781433105469}, {\"method\": \"loop\", \"foo\": \"foo\", \"n\": 7600, \"time\": 0.0015568733215332031}, {\"method\": \"loop\", \"foo\": \"foo\", \"n\": 8100, \"time\": 0.0015900135040283203}, {\"method\": \"loop\", \"foo\": \"foo\", \"n\": 8600, \"time\": 0.0021409988403320312}, {\"method\": \"loop\", \"foo\": \"foo\", \"n\": 9100, \"time\": 0.0027129650115966797}, {\"method\": \"loop\", \"foo\": \"foo\", \"n\": 9600, \"time\": 0.0027561187744140625}, {\"method\": \"loop\", \"foo\": \"foo_IO\", \"n\": 100, \"time\": 0.08153605461120605}, {\"method\": \"loop\", \"foo\": \"foo_IO\", \"n\": 600, \"time\": 0.31783509254455566}, {\"method\": \"loop\", \"foo\": \"foo_IO\", \"n\": 1100, \"time\": 0.6194419860839844}, {\"method\": \"loop\", \"foo\": \"foo_IO\", \"n\": 1600, \"time\": 0.8684618473052979}, {\"method\": \"loop\", \"foo\": \"foo_IO\", \"n\": 2100, \"time\": 1.3105027675628662}, {\"method\": \"loop\", \"foo\": \"foo_IO\", \"n\": 2600, \"time\": 1.3374419212341309}, {\"method\": \"loop\", \"foo\": \"foo_IO\", \"n\": 3100, \"time\": 1.9491069316864014}, {\"method\": \"loop\", \"foo\": \"foo_IO\", \"n\": 3600, \"time\": 1.8426711559295654}, {\"method\": \"loop\", \"foo\": \"foo_IO\", \"n\": 4100, \"time\": 2.4967620372772217}, {\"method\": \"loop\", \"foo\": \"foo_IO\", \"n\": 4600, \"time\": 2.3636231422424316}, {\"method\": \"loop\", \"foo\": \"foo_IO\", \"n\": 5100, \"time\": 3.0901639461517334}, {\"method\": \"loop\", \"foo\": \"foo_IO\", \"n\": 5600, \"time\": 3.1448090076446533}, {\"method\": \"loop\", \"foo\": \"foo_IO\", \"n\": 6100, \"time\": 3.644137144088745}, {\"method\": \"loop\", \"foo\": \"foo_IO\", \"n\": 6600, \"time\": 3.59538197517395}, {\"method\": \"loop\", \"foo\": \"foo_IO\", \"n\": 7100, \"time\": 4.382446765899658}, {\"method\": \"loop\", \"foo\": \"foo_IO\", \"n\": 7600, \"time\": 4.51059889793396}, {\"method\": \"loop\", \"foo\": \"foo_IO\", \"n\": 8100, \"time\": 4.8496057987213135}, {\"method\": \"loop\", \"foo\": \"foo_IO\", \"n\": 8600, \"time\": 4.615773916244507}, {\"method\": \"loop\", \"foo\": \"foo_IO\", \"n\": 9100, \"time\": 5.425705909729004}, {\"method\": \"loop\", \"foo\": \"foo_IO\", \"n\": 9600, \"time\": 5.690819978713989}, {\"method\": \"loopLog\", \"foo\": \"foo\", \"n\": 100, \"time\": 0.00014901161193847656}, {\"method\": \"loopLog\", \"foo\": \"foo\", \"n\": 600, \"time\": 0.0011568069458007812}, {\"method\": \"loopLog\", \"foo\": \"foo\", \"n\": 1100, \"time\": 0.002340078353881836}, {\"method\": \"loopLog\", \"foo\": \"foo\", \"n\": 1600, \"time\": 0.003278970718383789}, {\"method\": \"loopLog\", \"foo\": \"foo\", \"n\": 2100, \"time\": 0.004466056823730469}, {\"method\": \"loopLog\", \"foo\": \"foo\", \"n\": 2600, \"time\": 0.005352973937988281}, {\"method\": \"loopLog\", \"foo\": \"foo\", \"n\": 3100, \"time\": 0.0062711238861083984}, {\"method\": \"loopLog\", \"foo\": \"foo\", \"n\": 3600, \"time\": 0.007719278335571289}, {\"method\": \"loopLog\", \"foo\": \"foo\", \"n\": 4100, \"time\": 0.00992727279663086}, {\"method\": \"loopLog\", \"foo\": \"foo\", \"n\": 4600, \"time\": 0.010280847549438477}, {\"method\": \"loopLog\", \"foo\": \"foo\", \"n\": 5100, \"time\": 0.010870218276977539}, {\"method\": \"loopLog\", \"foo\": \"foo\", \"n\": 5600, \"time\": 0.011372804641723633}, {\"method\": \"loopLog\", \"foo\": \"foo\", \"n\": 6100, \"time\": 0.012415885925292969}, {\"method\": \"loopLog\", \"foo\": \"foo\", \"n\": 6600, \"time\": 0.013373851776123047}, {\"method\": \"loopLog\", \"foo\": \"foo\", \"n\": 7100, \"time\": 0.014763116836547852}, {\"method\": \"loopLog\", \"foo\": \"foo\", \"n\": 7600, \"time\": 0.016620159149169922}, {\"method\": \"loopLog\", \"foo\": \"foo\", \"n\": 8100, \"time\": 0.01786494255065918}, {\"method\": \"loopLog\", \"foo\": \"foo\", \"n\": 8600, \"time\": 0.02008199691772461}, {\"method\": \"loopLog\", \"foo\": \"foo\", \"n\": 9100, \"time\": 0.02127814292907715}, {\"method\": \"loopLog\", \"foo\": \"foo\", \"n\": 9600, \"time\": 0.02261209487915039}, {\"method\": \"nestedLoop\", \"foo\": \"foo\", \"n\": 100, \"time\": 0.0014941692352294922}, {\"method\": \"nestedLoop\", \"foo\": \"foo\", \"n\": 600, \"time\": 0.058754920959472656}, {\"method\": \"nestedLoop\", \"foo\": \"foo\", \"n\": 1100, \"time\": 0.18778610229492188}, {\"method\": \"nestedLoop\", \"foo\": \"foo\", \"n\": 1600, \"time\": 0.37906885147094727}, {\"method\": \"nestedLoop\", \"foo\": \"foo\", \"n\": 2100, \"time\": 0.6566951274871826}, {\"method\": \"nestedLoop\", \"foo\": \"foo\", \"n\": 2600, \"time\": 1.002432107925415}, {\"method\": \"nestedLoop\", \"foo\": \"foo\", \"n\": 3100, \"time\": 1.430429220199585}, {\"method\": \"nestedLoop\", \"foo\": \"foo\", \"n\": 3600, \"time\": 1.9189109802246094}, {\"method\": \"nestedLoop\", \"foo\": \"foo\", \"n\": 4100, \"time\": 2.481861114501953}, {\"method\": \"nestedLoop\", \"foo\": \"foo\", \"n\": 4600, \"time\": 3.122974395751953}, {\"method\": \"nestedLoop\", \"foo\": \"foo\", \"n\": 5100, \"time\": 3.8422510623931885}, {\"method\": \"nestedLoop\", \"foo\": \"foo\", \"n\": 5600, \"time\": 4.654331922531128}, {\"method\": \"nestedLoop\", \"foo\": \"foo\", \"n\": 6100, \"time\": 5.514336824417114}, {\"method\": \"nestedLoop\", \"foo\": \"foo\", \"n\": 6600, \"time\": 6.4394121170043945}, {\"method\": \"nestedLoop\", \"foo\": \"foo\", \"n\": 7100, \"time\": 7.525950193405151}, {\"method\": \"nestedLoop\", \"foo\": \"foo\", \"n\": 7600, \"time\": 8.57637619972229}, {\"method\": \"nestedLoop\", \"foo\": \"foo\", \"n\": 8100, \"time\": 9.76970386505127}, {\"method\": \"nestedLoop\", \"foo\": \"foo\", \"n\": 8600, \"time\": 10.94216012954712}, {\"method\": \"nestedLoop\", \"foo\": \"foo\", \"n\": 9100, \"time\": 12.33841609954834}, {\"method\": \"nestedLoop\", \"foo\": \"foo\", \"n\": 9600, \"time\": 13.721357822418213}]}}, {\"mode\": \"vega-lite\"});\n",
       "</script>"
      ],
      "text/plain": [
       "alt.Chart(...)"
      ]
     },
     "execution_count": 12,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "import altair as alt\n",
    "alt.renderers.enable('default')\n",
    "import pandas as pd\n",
    "\n",
    "# display all methods:\n",
    "# convert the <results> array to a pandas dataframe (a table-like structure)\n",
    "resultsPD = pd.DataFrame(results, columns=['method', 'foo', 'n', 'time'])\n",
    "\n",
    "# map the columns of resultsPD to x, y, and color\n",
    "alt.Chart(resultsPD).transform_calculate(\n",
    "    cat=\"datum.method + ' using ' + datum.foo\"\n",
    ").mark_line().encode(\n",
    "    x='n',\n",
    "    y=alt.Y('time', title='Time [sec]'),\n",
    "    color=alt.Y('cat:N', title='Algorithm'),\n",
    ").interactive()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 13,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>method</th>\n",
       "      <th>foo</th>\n",
       "      <th>n</th>\n",
       "      <th>time</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>0</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo</td>\n",
       "      <td>100</td>\n",
       "      <td>0.000022</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo</td>\n",
       "      <td>600</td>\n",
       "      <td>0.000115</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo</td>\n",
       "      <td>1100</td>\n",
       "      <td>0.000212</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>3</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo</td>\n",
       "      <td>1600</td>\n",
       "      <td>0.000329</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo</td>\n",
       "      <td>2100</td>\n",
       "      <td>0.000456</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>5</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo</td>\n",
       "      <td>2600</td>\n",
       "      <td>0.000631</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>6</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo</td>\n",
       "      <td>3100</td>\n",
       "      <td>0.000710</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>7</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo</td>\n",
       "      <td>3600</td>\n",
       "      <td>0.001056</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>8</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo</td>\n",
       "      <td>4100</td>\n",
       "      <td>0.000920</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>9</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo</td>\n",
       "      <td>4600</td>\n",
       "      <td>0.000984</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>10</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo</td>\n",
       "      <td>5100</td>\n",
       "      <td>0.001076</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>11</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo</td>\n",
       "      <td>5600</td>\n",
       "      <td>0.001218</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>12</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo</td>\n",
       "      <td>6100</td>\n",
       "      <td>0.001297</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>13</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo</td>\n",
       "      <td>6600</td>\n",
       "      <td>0.001538</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>14</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo</td>\n",
       "      <td>7100</td>\n",
       "      <td>0.001701</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>15</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo</td>\n",
       "      <td>7600</td>\n",
       "      <td>0.001557</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>16</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo</td>\n",
       "      <td>8100</td>\n",
       "      <td>0.001590</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>17</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo</td>\n",
       "      <td>8600</td>\n",
       "      <td>0.002141</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>18</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo</td>\n",
       "      <td>9100</td>\n",
       "      <td>0.002713</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>19</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo</td>\n",
       "      <td>9600</td>\n",
       "      <td>0.002756</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>20</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo_IO</td>\n",
       "      <td>100</td>\n",
       "      <td>0.081536</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>21</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo_IO</td>\n",
       "      <td>600</td>\n",
       "      <td>0.317835</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>22</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo_IO</td>\n",
       "      <td>1100</td>\n",
       "      <td>0.619442</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>23</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo_IO</td>\n",
       "      <td>1600</td>\n",
       "      <td>0.868462</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>24</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo_IO</td>\n",
       "      <td>2100</td>\n",
       "      <td>1.310503</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>25</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo_IO</td>\n",
       "      <td>2600</td>\n",
       "      <td>1.337442</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>26</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo_IO</td>\n",
       "      <td>3100</td>\n",
       "      <td>1.949107</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>27</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo_IO</td>\n",
       "      <td>3600</td>\n",
       "      <td>1.842671</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>28</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo_IO</td>\n",
       "      <td>4100</td>\n",
       "      <td>2.496762</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>29</th>\n",
       "      <td>loop</td>\n",
       "      <td>foo_IO</td>\n",
       "      <td>4600</td>\n",
       "      <td>2.363623</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>...</th>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>50</th>\n",
       "      <td>loopLog</td>\n",
       "      <td>foo</td>\n",
       "      <td>5100</td>\n",
       "      <td>0.010870</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>51</th>\n",
       "      <td>loopLog</td>\n",
       "      <td>foo</td>\n",
       "      <td>5600</td>\n",
       "      <td>0.011373</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>52</th>\n",
       "      <td>loopLog</td>\n",
       "      <td>foo</td>\n",
       "      <td>6100</td>\n",
       "      <td>0.012416</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>53</th>\n",
       "      <td>loopLog</td>\n",
       "      <td>foo</td>\n",
       "      <td>6600</td>\n",
       "      <td>0.013374</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>54</th>\n",
       "      <td>loopLog</td>\n",
       "      <td>foo</td>\n",
       "      <td>7100</td>\n",
       "      <td>0.014763</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>55</th>\n",
       "      <td>loopLog</td>\n",
       "      <td>foo</td>\n",
       "      <td>7600</td>\n",
       "      <td>0.016620</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>56</th>\n",
       "      <td>loopLog</td>\n",
       "      <td>foo</td>\n",
       "      <td>8100</td>\n",
       "      <td>0.017865</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>57</th>\n",
       "      <td>loopLog</td>\n",
       "      <td>foo</td>\n",
       "      <td>8600</td>\n",
       "      <td>0.020082</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>58</th>\n",
       "      <td>loopLog</td>\n",
       "      <td>foo</td>\n",
       "      <td>9100</td>\n",
       "      <td>0.021278</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>59</th>\n",
       "      <td>loopLog</td>\n",
       "      <td>foo</td>\n",
       "      <td>9600</td>\n",
       "      <td>0.022612</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>60</th>\n",
       "      <td>nestedLoop</td>\n",
       "      <td>foo</td>\n",
       "      <td>100</td>\n",
       "      <td>0.001494</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>61</th>\n",
       "      <td>nestedLoop</td>\n",
       "      <td>foo</td>\n",
       "      <td>600</td>\n",
       "      <td>0.058755</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>62</th>\n",
       "      <td>nestedLoop</td>\n",
       "      <td>foo</td>\n",
       "      <td>1100</td>\n",
       "      <td>0.187786</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>63</th>\n",
       "      <td>nestedLoop</td>\n",
       "      <td>foo</td>\n",
       "      <td>1600</td>\n",
       "      <td>0.379069</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>64</th>\n",
       "      <td>nestedLoop</td>\n",
       "      <td>foo</td>\n",
       "      <td>2100</td>\n",
       "      <td>0.656695</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>65</th>\n",
       "      <td>nestedLoop</td>\n",
       "      <td>foo</td>\n",
       "      <td>2600</td>\n",
       "      <td>1.002432</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>66</th>\n",
       "      <td>nestedLoop</td>\n",
       "      <td>foo</td>\n",
       "      <td>3100</td>\n",
       "      <td>1.430429</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>67</th>\n",
       "      <td>nestedLoop</td>\n",
       "      <td>foo</td>\n",
       "      <td>3600</td>\n",
       "      <td>1.918911</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>68</th>\n",
       "      <td>nestedLoop</td>\n",
       "      <td>foo</td>\n",
       "      <td>4100</td>\n",
       "      <td>2.481861</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>69</th>\n",
       "      <td>nestedLoop</td>\n",
       "      <td>foo</td>\n",
       "      <td>4600</td>\n",
       "      <td>3.122974</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>70</th>\n",
       "      <td>nestedLoop</td>\n",
       "      <td>foo</td>\n",
       "      <td>5100</td>\n",
       "      <td>3.842251</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>71</th>\n",
       "      <td>nestedLoop</td>\n",
       "      <td>foo</td>\n",
       "      <td>5600</td>\n",
       "      <td>4.654332</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>72</th>\n",
       "      <td>nestedLoop</td>\n",
       "      <td>foo</td>\n",
       "      <td>6100</td>\n",
       "      <td>5.514337</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>73</th>\n",
       "      <td>nestedLoop</td>\n",
       "      <td>foo</td>\n",
       "      <td>6600</td>\n",
       "      <td>6.439412</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>74</th>\n",
       "      <td>nestedLoop</td>\n",
       "      <td>foo</td>\n",
       "      <td>7100</td>\n",
       "      <td>7.525950</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>75</th>\n",
       "      <td>nestedLoop</td>\n",
       "      <td>foo</td>\n",
       "      <td>7600</td>\n",
       "      <td>8.576376</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>76</th>\n",
       "      <td>nestedLoop</td>\n",
       "      <td>foo</td>\n",
       "      <td>8100</td>\n",
       "      <td>9.769704</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>77</th>\n",
       "      <td>nestedLoop</td>\n",
       "      <td>foo</td>\n",
       "      <td>8600</td>\n",
       "      <td>10.942160</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>78</th>\n",
       "      <td>nestedLoop</td>\n",
       "      <td>foo</td>\n",
       "      <td>9100</td>\n",
       "      <td>12.338416</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>79</th>\n",
       "      <td>nestedLoop</td>\n",
       "      <td>foo</td>\n",
       "      <td>9600</td>\n",
       "      <td>13.721358</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "<p>80 rows × 4 columns</p>\n",
       "</div>"
      ],
      "text/plain": [
       "        method     foo     n       time\n",
       "0         loop     foo   100   0.000022\n",
       "1         loop     foo   600   0.000115\n",
       "2         loop     foo  1100   0.000212\n",
       "3         loop     foo  1600   0.000329\n",
       "4         loop     foo  2100   0.000456\n",
       "5         loop     foo  2600   0.000631\n",
       "6         loop     foo  3100   0.000710\n",
       "7         loop     foo  3600   0.001056\n",
       "8         loop     foo  4100   0.000920\n",
       "9         loop     foo  4600   0.000984\n",
       "10        loop     foo  5100   0.001076\n",
       "11        loop     foo  5600   0.001218\n",
       "12        loop     foo  6100   0.001297\n",
       "13        loop     foo  6600   0.001538\n",
       "14        loop     foo  7100   0.001701\n",
       "15        loop     foo  7600   0.001557\n",
       "16        loop     foo  8100   0.001590\n",
       "17        loop     foo  8600   0.002141\n",
       "18        loop     foo  9100   0.002713\n",
       "19        loop     foo  9600   0.002756\n",
       "20        loop  foo_IO   100   0.081536\n",
       "21        loop  foo_IO   600   0.317835\n",
       "22        loop  foo_IO  1100   0.619442\n",
       "23        loop  foo_IO  1600   0.868462\n",
       "24        loop  foo_IO  2100   1.310503\n",
       "25        loop  foo_IO  2600   1.337442\n",
       "26        loop  foo_IO  3100   1.949107\n",
       "27        loop  foo_IO  3600   1.842671\n",
       "28        loop  foo_IO  4100   2.496762\n",
       "29        loop  foo_IO  4600   2.363623\n",
       "..         ...     ...   ...        ...\n",
       "50     loopLog     foo  5100   0.010870\n",
       "51     loopLog     foo  5600   0.011373\n",
       "52     loopLog     foo  6100   0.012416\n",
       "53     loopLog     foo  6600   0.013374\n",
       "54     loopLog     foo  7100   0.014763\n",
       "55     loopLog     foo  7600   0.016620\n",
       "56     loopLog     foo  8100   0.017865\n",
       "57     loopLog     foo  8600   0.020082\n",
       "58     loopLog     foo  9100   0.021278\n",
       "59     loopLog     foo  9600   0.022612\n",
       "60  nestedLoop     foo   100   0.001494\n",
       "61  nestedLoop     foo   600   0.058755\n",
       "62  nestedLoop     foo  1100   0.187786\n",
       "63  nestedLoop     foo  1600   0.379069\n",
       "64  nestedLoop     foo  2100   0.656695\n",
       "65  nestedLoop     foo  2600   1.002432\n",
       "66  nestedLoop     foo  3100   1.430429\n",
       "67  nestedLoop     foo  3600   1.918911\n",
       "68  nestedLoop     foo  4100   2.481861\n",
       "69  nestedLoop     foo  4600   3.122974\n",
       "70  nestedLoop     foo  5100   3.842251\n",
       "71  nestedLoop     foo  5600   4.654332\n",
       "72  nestedLoop     foo  6100   5.514337\n",
       "73  nestedLoop     foo  6600   6.439412\n",
       "74  nestedLoop     foo  7100   7.525950\n",
       "75  nestedLoop     foo  7600   8.576376\n",
       "76  nestedLoop     foo  8100   9.769704\n",
       "77  nestedLoop     foo  8600  10.942160\n",
       "78  nestedLoop     foo  9100  12.338416\n",
       "79  nestedLoop     foo  9600  13.721358\n",
       "\n",
       "[80 rows x 4 columns]"
      ]
     },
     "execution_count": 13,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "# display the data as a table:\n",
    "resultsPD"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "feel free to ignore the following..."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 14,
   "metadata": {},
   "outputs": [],
   "source": [
    "# helper to be able to tun SQL against pandas data frames:\n",
    "import pandasql as pdsql\n",
    "pysql = lambda q: pdsql.sqldf(q, globals())\n",
    "import sqlparse\n",
    "\n",
    "def queryFrame(q, _display=False):\n",
    "    result = pysql(q)\n",
    "    if _display:\n",
    "        display(result)\n",
    "    return result"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 15,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>n</th>\n",
       "      <th>improvementFactor</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>0</th>\n",
       "      <td>100</td>\n",
       "      <td>3717.250000</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>600</td>\n",
       "      <td>2765.761411</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>1100</td>\n",
       "      <td>2922.528684</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>3</th>\n",
       "      <td>1600</td>\n",
       "      <td>2641.474257</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>2100</td>\n",
       "      <td>2874.815377</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>5</th>\n",
       "      <td>2600</td>\n",
       "      <td>2119.243672</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>6</th>\n",
       "      <td>3100</td>\n",
       "      <td>2744.258812</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>7</th>\n",
       "      <td>3600</td>\n",
       "      <td>1745.026643</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>8</th>\n",
       "      <td>4100</td>\n",
       "      <td>2714.406169</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>9</th>\n",
       "      <td>4600</td>\n",
       "      <td>2401.587694</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>10</th>\n",
       "      <td>5100</td>\n",
       "      <td>2871.944826</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>11</th>\n",
       "      <td>5600</td>\n",
       "      <td>2582.279757</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>12</th>\n",
       "      <td>6100</td>\n",
       "      <td>2809.156221</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>13</th>\n",
       "      <td>6600</td>\n",
       "      <td>2338.003876</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>14</th>\n",
       "      <td>7100</td>\n",
       "      <td>2576.578918</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>15</th>\n",
       "      <td>7600</td>\n",
       "      <td>2897.216386</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>16</th>\n",
       "      <td>8100</td>\n",
       "      <td>3050.040636</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>17</th>\n",
       "      <td>8600</td>\n",
       "      <td>2155.897439</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>18</th>\n",
       "      <td>9100</td>\n",
       "      <td>1999.917392</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>19</th>\n",
       "      <td>9600</td>\n",
       "      <td>2064.794896</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "</div>"
      ],
      "text/plain": [
       "       n  improvementFactor\n",
       "0    100        3717.250000\n",
       "1    600        2765.761411\n",
       "2   1100        2922.528684\n",
       "3   1600        2641.474257\n",
       "4   2100        2874.815377\n",
       "5   2600        2119.243672\n",
       "6   3100        2744.258812\n",
       "7   3600        1745.026643\n",
       "8   4100        2714.406169\n",
       "9   4600        2401.587694\n",
       "10  5100        2871.944826\n",
       "11  5600        2582.279757\n",
       "12  6100        2809.156221\n",
       "13  6600        2338.003876\n",
       "14  7100        2576.578918\n",
       "15  7600        2897.216386\n",
       "16  8100        3050.040636\n",
       "17  8600        2155.897439\n",
       "18  9100        1999.917392\n",
       "19  9600        2064.794896"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "# calculate performance difference of different methods as a facor:\n",
    "improvementFactor = queryFrame(\"\"\"\n",
    "WITH justfoo AS (\n",
    "    SELECT *\n",
    "    FROM resultsPD\n",
    "    WHERE method='loop' AND foo='foo'),\n",
    "justRead AS (\n",
    "    SELECT *\n",
    "    FROM resultsPD\n",
    "    WHERE method='loop' AND foo='foo_IO')\n",
    "\n",
    "SELECT L.n, R.time/L.time as improvementFactor\n",
    "FROM justfoo L JOIN justRead R\n",
    "ON L.n==R.n\n",
    ";\n",
    "\"\"\", True)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 16,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "\n",
       "<div id=\"altair-viz-5\"></div>\n",
       "<script type=\"text/javascript\">\n",
       "  (function(spec, embedOpt){\n",
       "    const outputDiv = document.getElementById(\"altair-viz-5\");\n",
       "    const paths = {\n",
       "      \"vega\": \"https://cdn.jsdelivr.net/npm//vega@5?noext\",\n",
       "      \"vega-lib\": \"https://cdn.jsdelivr.net/npm//vega-lib?noext\",\n",
       "      \"vega-lite\": \"https://cdn.jsdelivr.net/npm//vega-lite@4.0.0?noext\",\n",
       "      \"vega-embed\": \"https://cdn.jsdelivr.net/npm//vega-embed@6?noext\",\n",
       "    };\n",
       "\n",
       "    function loadScript(lib) {\n",
       "      return new Promise(function(resolve, reject) {\n",
       "        var s = document.createElement('script');\n",
       "        s.src = paths[lib];\n",
       "        s.async = true;\n",
       "        s.onload = () => resolve(paths[lib]);\n",
       "        s.onerror = () => reject(`Error loading script: ${paths[lib]}`);\n",
       "        document.getElementsByTagName(\"head\")[0].appendChild(s);\n",
       "      });\n",
       "    }\n",
       "\n",
       "    function showError(err) {\n",
       "      outputDiv.innerHTML = `<div class=\"error\" style=\"color:red;\">${err}</div>`;\n",
       "      throw err;\n",
       "    }\n",
       "\n",
       "    function displayChart(vegaEmbed) {\n",
       "      vegaEmbed(outputDiv, spec, embedOpt)\n",
       "        .catch(err => showError(`Javascript Error: ${err.message}<br>This usually means there's a typo in your chart specification. See the javascript console for the full traceback.`));\n",
       "    }\n",
       "\n",
       "    if(typeof define === \"function\" && define.amd) {\n",
       "      requirejs.config({paths});\n",
       "      require([\"vega-embed\"], displayChart, err => showError(`Error loading script: ${err.message}`));\n",
       "    } else if (typeof vegaEmbed === \"function\") {\n",
       "      displayChart(vegaEmbed);\n",
       "    } else {\n",
       "      loadScript(\"vega\")\n",
       "        .then(() => loadScript(\"vega-lite\"))\n",
       "        .then(() => loadScript(\"vega-embed\"))\n",
       "        .catch(showError)\n",
       "        .then(() => displayChart(vegaEmbed));\n",
       "    }\n",
       "  })({\"config\": {\"view\": {\"continuousWidth\": 400, \"continuousHeight\": 300}}, \"data\": {\"name\": \"data-65d288d1a0ffe9d629e4785b3f178146\"}, \"mark\": \"line\", \"encoding\": {\"x\": {\"type\": \"quantitative\", \"field\": \"n\"}, \"y\": {\"type\": \"quantitative\", \"field\": \"improvementFactor\", \"title\": \"improvement factor\"}}, \"$schema\": \"https://vega.github.io/schema/vega-lite/v4.0.0.json\", \"datasets\": {\"data-65d288d1a0ffe9d629e4785b3f178146\": [{\"n\": 100, \"improvementFactor\": 3717.25}, {\"n\": 600, \"improvementFactor\": 2765.761410788382}, {\"n\": 1100, \"improvementFactor\": 2922.5286839145106}, {\"n\": 1600, \"improvementFactor\": 2641.474256707759}, {\"n\": 2100, \"improvementFactor\": 2874.8153765690377}, {\"n\": 2600, \"improvementFactor\": 2119.2436720816017}, {\"n\": 3100, \"improvementFactor\": 2744.2588116817724}, {\"n\": 3600, \"improvementFactor\": 1745.026642582976}, {\"n\": 4100, \"improvementFactor\": 2714.4061689994815}, {\"n\": 4600, \"improvementFactor\": 2401.5876937984494}, {\"n\": 5100, \"improvementFactor\": 2871.9448260580543}, {\"n\": 5600, \"improvementFactor\": 2582.2797572435397}, {\"n\": 6100, \"improvementFactor\": 2809.1562212828526}, {\"n\": 6600, \"improvementFactor\": 2338.0038759689924}, {\"n\": 7100, \"improvementFactor\": 2576.578917858144}, {\"n\": 7600, \"improvementFactor\": 2897.2163859111793}, {\"n\": 8100, \"improvementFactor\": 3050.0406357774777}, {\"n\": 8600, \"improvementFactor\": 2155.897438752784}, {\"n\": 9100, \"improvementFactor\": 1999.91739168644}, {\"n\": 9600, \"improvementFactor\": 2064.794896193772}]}}, {\"mode\": \"vega-lite\"});\n",
       "</script>"
      ],
      "text/plain": [
       "alt.Chart(...)"
      ]
     },
     "execution_count": 16,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "# draw the improvement factor:\n",
    "# this graph shows the \"slow donw\" of the method without I/O over the one with I/O\n",
    "alt.Chart(improvementFactor).mark_line().encode(\n",
    "    x='n',\n",
    "    y=alt.Y('improvementFactor', title='improvement factor')\n",
    ")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.7.4"
  },
  "varInspector": {
   "cols": {
    "lenName": 16,
    "lenType": 16,
    "lenVar": 40
   },
   "kernels_config": {
    "python": {
     "delete_cmd_postfix": "",
     "delete_cmd_prefix": "del ",
     "library": "var_list.py",
     "varRefreshCmd": "print(var_dic_list())"
    },
    "r": {
     "delete_cmd_postfix": ") ",
     "delete_cmd_prefix": "rm(",
     "library": "var_list.r",
     "varRefreshCmd": "cat(var_dic_list()) "
    }
   },
   "types_to_exclude": [
    "module",
    "function",
    "builtin_function_or_method",
    "instance",
    "_Feature"
   ],
   "window_display": false
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
